"""
audio_case_grade: A Python library for analyzing and grading audio cases.
"""

from .clean import clean, get_abbreviations_csv
from .core import get_score, hello
from .grade import get_grade
from .keybank import get_keybank, get_keybank_csv
from .keyword import count_keywords, get_keywords
from .types import (
    Assessments,
    Case,
    Histories,
    Metrics,
    Objectives,
    Plans,
    Score,
    Soap,
    Totals,
    Transcriber,
    Transcript,
)

__all__ = [
    "clean",
    "count_keywords",
    "hello",
    "get_abbreviations_csv",
    "get_grade",
    "get_keybank",
    "get_keybank_csv",
    "get_keywords",
    "get_score",
    "Assessments",
    "Case",
    "Histories",
    "Metrics",
    "Objectives",
    "Plans",
    "Score",
    "Soap",
    "Totals",
    "Transcriber",
    "Transcript",
]



"""Type definitions to support core"""

# pylint: disable=too-few-public-methods

from enum import Enum
from typing import Dict, List, Optional, cast, get_type_hints

from pydantic import BaseModel


class Transcriber(str, Enum):
    """Enum for transcriber"""

    WHISPER = "OpenAI Whisper"
    DEEPGRAM = "Deepgram Voice Agent"


class Transcript(BaseModel):
    """Type definition for a transcript"""

    type: Transcriber
    raw: str
    clean: str = ""

    @property
    def word_count(self) -> int:
        """Word count of the clean transcript"""
        return len(self.clean.split())


class Case(BaseModel):
    """ "Type definition for a case"""

    code: str
    """ICD-10 code"""

    system: str
    """Body System"""

    name: str


class Metrics(BaseModel):
    """Type definition for metrics"""

    count: int
    """Actual amount of keywords used"""

    expected: int
    """Expected keyword count based on the keybank"""

    sequence: Dict[str, int]
    """Sequence of keywords used"""

    used: List[str]
    """List of keywords used"""

    @property
    def word_count(self) -> int:
        """Word count of keywords used, keywords may contain multiple words"""

        used = " ".join(self.used)
        return len(used.split())

    @property
    def percentage(self) -> float:
        """Percentage of keywords used"""

        return round((self.count / self.expected), 4) * 100


class Totals(BaseModel):
    """Type definition for totals"""

    count: int
    """Actual amount of keywords used"""

    expected: int
    """Expected keyword count based on the keybank"""

    word_count: int
    """Word count of keywords used, keywords may contain multiple words"""

    @property
    def percentage(self) -> float:
        """Percentage of keywords used"""

        return round((self.count / self.expected), 4) * 100


class TotalsRollupMixIn:
    """Mix-in class to calculate totals for all properties of type Metrics."""

    def calculate_totals(self) -> Totals:
        """Calculate totals for all properties of type Metrics."""
        total_count = 0
        total_expected = 0
        total_word_count = 0

        # Iterate over all attributes of the class
        for name, field_type in get_type_hints(type(self)).items():
            if field_type in (Optional[Metrics], Metrics):
                value = getattr(self, name)
                if value and value is not None:
                    metrics = cast(Metrics, value)
                    total_count += metrics.count
                    total_expected += metrics.expected
                    total_word_count += metrics.word_count

        return Totals(
            count=total_count,
            expected=total_expected,
            word_count=total_word_count,
        )


class Histories(BaseModel, TotalsRollupMixIn):
    """Type definition for histories"""

    cc: Optional[Metrics] = None
    """Cardinal Complaint"""
    hpi: Optional[Metrics] = None
    """History of Present Illness"""
    ros: Optional[Metrics] = None
    """Review of System"""
    meds: Optional[Metrics] = None
    """Medications"""

    @property
    def totals(self) -> Totals:
        """Totals for the histories"""
        return self.calculate_totals()


class Objectives(BaseModel, TotalsRollupMixIn):
    """Type definition for objectives"""

    vitals: Optional[Metrics] = None
    """Vital Signs"""
    gen: Optional[Metrics] = None
    """General Appearance"""
    pe: Optional[Metrics] = None
    """Physical Exam"""
    dl: Optional[Metrics] = None
    """Diagnostic Labs"""
    di: Optional[Metrics] = None
    """Diagnostic Imaging"""

    @property
    def totals(self) -> Totals:
        """Totals for the objectives"""
        return self.calculate_totals()


class Assessments(BaseModel, TotalsRollupMixIn):
    """ "Type definition for assessments"""

    dx: Optional[Metrics] = None
    """Diagnosis"""
    ddx: Optional[Metrics] = None
    """Differentials"""

    @property
    def totals(self) -> Totals:
        """Totals for the assessments"""
        return self.calculate_totals()


class Plans(BaseModel, TotalsRollupMixIn):
    """Type definition for plans"""

    tx: Optional[Metrics] = None
    """Treatment"""
    consults: Optional[Metrics] = None
    """Physician and Specialty Consultations"""
    interventions: Optional[Metrics] = None
    """Surgeries and Special Procedures"""

    @property
    def totals(self) -> Totals:
        """Totals for the plans"""
        return self.calculate_totals()


class Foundations(BaseModel, TotalsRollupMixIn):
    """Type definition for foundations"""

    root: Optional[Metrics] = None
    """Foundation"""

    @property
    def totals(self) -> Totals:
        """Totals for the foundations"""
        return self.calculate_totals()


class Soap(BaseModel):
    """Type definition for Subjective Objective Assessment Plan (SOAP)"""

    histories: Histories
    objectives: Objectives
    assessments: Assessments
    plans: Plans
    foundations: Foundations

    @property
    def totals(self) -> Totals:
        """Totals for the SOAP"""
        components_totals = [
            self.histories.totals,
            self.objectives.totals,
            self.assessments.totals,
            self.plans.totals,
            self.foundations.totals,
        ]

        total_count = sum(t.count for t in components_totals)
        total_expected = sum(t.expected for t in components_totals)
        total_word_count = sum(t.word_count for t in components_totals)

        return Totals(
            count=total_count, expected=total_expected, word_count=total_word_count
        )


class Score(BaseModel):
    """Type definition for a score"""

    lexical_density: float
    correct: Soap
    """Correct Medical Terminology"""
    wrong: Soap
    """Wrong Medical Terminology"""



"""Testing for types module"""

from audio_case_grade import Metrics, Transcriber, Transcript


def test_transcript_word_count():
    """Test transcript type"""

    transcript = Transcript(
        type=Transcriber.DEEPGRAM,
        raw="This is the raw text",
        clean="clean text",
    )

    assert transcript.word_count == 2


def test_metrics_word_count():
    """Test metrics word count"""

    metrics = Metrics(
        count=2,
        expected=3,
        sequence={"test": 1, "test2": 2, "test3": -1},
        used=["test", "test2"],
    )

    assert metrics.word_count == 2


def test_metrics_percentage():
    """Test metrics percentage"""

    metrics = Metrics(
        count=2,
        expected=3,
        sequence={"test": 1, "test2": 2, "test3": -1},
        used=["test", "test2"],
    )

    assert metrics.percentage == 66.67



"""Keyword algorithm module"""

import re
from typing import Optional, Tuple

from pandas import DataFrame, Series

from .keybank import get_keybank
from .types import (
    Assessments,
    Case,
    Foundations,
    Histories,
    Metrics,
    Objectives,
    Plans,
    Soap,
)


def count_keywords(text: str, keybank: list[str]) -> Optional[Metrics]:
    """Get the keyword count and other metrics for a piece of text"""

    expected = len(keybank)

    if expected == 0:
        return None

    keyword_sequence: dict[str, int] = {}
    keywords_used: list[str] = []

    count = 0
    not_found = -1

    for keyword in keybank:
        if keyword in text:
            keywords_used.append(keyword)
            count += 1
            match = re.search(keyword, text)
            keyword_sequence[keyword] = match.start() + 1 if match else not_found
        else:
            keyword_sequence[keyword] = not_found

    return Metrics(
        count=count,
        expected=expected,
        sequence=keyword_sequence,
        used=keywords_used,
    )


def _histories(text: str, keybank: Series) -> Histories:
    """Get the history metrics for a transcript"""

    cc = count_keywords(text, keybank.iloc[0])
    hpi = count_keywords(text, keybank.iloc[1])
    ros = count_keywords(text, keybank.iloc[2])
    meds = count_keywords(text, keybank.iloc[3])

    return Histories(cc=cc, hpi=hpi, ros=ros, meds=meds)


def _objectives(text: str, keybank: Series) -> Objectives:
    """Get the objectives metrics for a transcript"""

    vitals = count_keywords(text, keybank.iloc[4])
    gen = count_keywords(text, keybank.iloc[5])
    pe = count_keywords(text, keybank.iloc[6])
    dl = count_keywords(text, keybank.iloc[7])
    di = count_keywords(text, keybank.iloc[8])

    return Objectives(vitals=vitals, gen=gen, pe=pe, dl=dl, di=di)


def _assessments(text: str, keybank: Series) -> Assessments:
    """Get the assessment metrics for a transcript"""

    dx = count_keywords(text, keybank.iloc[9])
    ddx = count_keywords(text, keybank.iloc[10])

    return Assessments(dx=dx, ddx=ddx)


def _plans(text: str, keybank: Series) -> Plans:
    """Get the plan metrics for a transcript"""

    tx = count_keywords(text, keybank.iloc[11])
    consults = count_keywords(text, keybank.iloc[12])
    interventions = count_keywords(text, keybank.iloc[13])

    return Plans(tx=tx, consults=consults, interventions=interventions)


def _foundations(text: str, keybank: Series) -> Foundations:
    """Get the assessment metrics for a transcript"""

    root = count_keywords(text, keybank.iloc[14])

    return Foundations(root=root)


def _soap(
    text: str, correct_keybank: Series, wrong_keybank: Series
) -> Tuple[Soap, Soap]:
    """calculate all metrics of the SOAP note flow and return the data frames for each section"""

    correct = Soap(
        histories=_histories(text, correct_keybank),
        objectives=_objectives(text, correct_keybank),
        assessments=_assessments(text, correct_keybank),
        plans=_plans(text, correct_keybank),
        foundations=_foundations(text, correct_keybank),
    )

    wrong = Soap(
        histories=_histories(text, wrong_keybank),
        objectives=_objectives(text, wrong_keybank),
        assessments=_assessments(text, wrong_keybank),
        plans=_plans(text, wrong_keybank),
        foundations=_foundations(text, wrong_keybank),
    )

    return correct, wrong


def get_keywords(
    text: str, case: Case, keybank_override: Optional[DataFrame] = None
) -> tuple[Soap, Soap]:
    """main function for keyword algorithm, applies above functions to all student transcripts"""

    (correct_keybank, wrong_keybank) = get_keybank(case, keybank_override)
    (correct_soap, wrong_soap) = _soap(
        text,
        correct_keybank,
        wrong_keybank,
    )

    return correct_soap, wrong_soap



"""Top level display objects like grade and assignments for student"""


def get_grade(lexical_density: float) -> str:
    """Grade a score with a weighted grade on a scale of 100"""
    if lexical_density >= 40:
        return "Advanced"
    if 30 <= lexical_density < 40:
        return "Proficient"
    if 20 <= lexical_density < 30:
        return "Average"

    return "Low"



"""Core module for grading medical audio cases"""

from typing import Optional

from pandas import DataFrame

from .keyword import get_keywords
from .types import Case, Score, Transcript


def hello() -> str:
    """Welcome message for the package"""

    return "Welcome to audio-case-grade!"


def get_score(
    transcript: Transcript, case: Case, keybank_override: Optional[DataFrame] = None
) -> Score:
    """Score an audio case transcript"""

    correct_soap, wrong_soap = get_keywords(transcript.clean, case, keybank_override)

    lexical_density = (
        round((correct_soap.totals.word_count / transcript.word_count), 4) * 100
    )

    return Score(
        lexical_density=lexical_density,
        correct=correct_soap,
        wrong=wrong_soap,
    )



"""Testing for keyword module"""

from audio_case_grade import (
    Case,
    Transcriber,
    Transcript,
    clean,
    count_keywords,
    get_keywords,
)


def test_count_keywords():
    """Test count keywords function"""

    metrics = count_keywords(
        "This is a banana and apple test", ["banana", "apple", "kiwi"]
    )

    assert metrics is not None
    assert metrics.count == 2
    assert metrics.expected == 3
    assert metrics.sequence == {"banana": 11, "apple": 22, "kiwi": -1}
    assert metrics.used == ["banana", "apple"]


def test_get_keywords():
    """Test get keywords function Correct Terminology Count"""

    text = "hello world chest pain two months fatigued ago"
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )

    correct_soap, wrong_soap = get_keywords(text, case)

    assert correct_soap is not None
    assert correct_soap.totals.count == 3
    assert correct_soap.totals.expected == 52
    assert correct_soap.totals.word_count == 5

    assert wrong_soap is not None
    assert wrong_soap.totals.count == 0
    assert wrong_soap.totals.expected == 34
    assert wrong_soap.totals.word_count == 0


def test_get_keywords_wrong():
    """Test get keywords function Wrong Terminology Count"""

    text = "hello world chest pain two weeks palpitations pre syncope"
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )

    correct_soap, wrong_soap = get_keywords(text, case)

    assert correct_soap is not None
    assert correct_soap.totals.count == 1
    assert correct_soap.totals.expected == 52
    assert correct_soap.totals.word_count == 2

    assert wrong_soap is not None
    assert wrong_soap.totals.count == 3
    assert wrong_soap.totals.expected == 34
    assert wrong_soap.totals.word_count == 5


def test_get_keywords_percentage():
    """Test get keywords percentage"""

    text = "hello world chest pain two weeks palpitations pre syncope"
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )

    correct_soap, wrong_soap = get_keywords(text, case)

    assert correct_soap is not None
    assert correct_soap.totals.percentage == 1.92

    assert wrong_soap is not None
    assert wrong_soap.totals.percentage == 8.82


def test_get_keywords_usage():
    """Test get keywords usage"""

    stem_hpi = " last known well time two hours incontinence"
    stem_pe = " stoke severity moderate"
    stem_dx = " cerebral infarction aca"
    stem_tx = " asa seizure precautions dvt prophylaxis"
    stem_found = " acute ischemic stroke aca symptoms"
    text = stem_hpi + stem_pe + stem_dx + stem_tx + stem_found

    transcript = Transcript(
        type=Transcriber.DEEPGRAM, raw=text, word_count=len(text.split())
    )
    clean_txt = clean(transcript)
    case = Case(
        code="I63.43",
        system="Neruo",
        name="cerebral infarction aca < 3 hrs",
    )

    correct_soap, wrong_soap = get_keywords(clean_txt.clean, case)

    assert correct_soap is not None
    assert wrong_soap is not None
    assert sorted(correct_soap.histories.hpi.used) == sorted(
        ["last know well time", "two hour", "incontinence"]
    )
    assert correct_soap.objectives.pe.used == ["stoke severity moderate"]
    assert correct_soap.assessments.dx.used == ["cerebral infarction aca"]
    assert sorted(correct_soap.plans.tx.used) == sorted(
        ["asa", "seizure precaution", "deep venous thrombosis prophylaxis"]
    )
    assert sorted(correct_soap.foundations.root.used) == sorted(
        ["acute", "ischemic stroke", "aca", "symptom"]
    )


def test_get_keywords_sequence():
    """Test get keywords Sequence"""

    stem_hpi = " last known well time two hours incontinence"
    stem_pe = " stoke severity moderate"
    stem_dx = " cerebral infarction aca"
    stem_tx = " asa seizure precautions dvt prophylaxis"
    stem_found = " acute ischemic stroke aca symptoms"
    text = stem_hpi + stem_pe + stem_dx + stem_tx + stem_found

    transcript = Transcript(
        type=Transcriber.DEEPGRAM, raw=text, word_count=len(text.split())
    )
    clean_txt = clean(transcript)
    case = Case(
        code="I63.43",
        system="Neruo",
        name="cerebral infarction aca < 3 hrs",
    )

    correct_soap, wrong_soap = get_keywords(clean_txt.clean, case)

    assert correct_soap is not None
    assert wrong_soap is not None
    assert correct_soap.histories.hpi.sequence == dict(
        {"last know well time": 1, "two hour": 21, "incontinence": 30}
    )
    assert correct_soap.objectives.pe.sequence == dict({"stoke severity moderate": 43})
    assert correct_soap.assessments.dx.sequence == dict({"cerebral infarction aca": 67})
    assert correct_soap.plans.tx.sequence == dict(
        {
            "asa": 91,
            "seizure precaution": 95,
            "deep venous thrombosis prophylaxis": 114,
            "insulin slide scale": -1,
        }
    )

    assert correct_soap.foundations.root.sequence == dict(
        {"acute": 149, "ischemic stroke": 155, "aca": 87, "symptom": 175}
    )


def test_get_keywords_soap_count():
    """Test get keywords function for SOAP"""

    def create_case() -> str:
        stem_cc = "stroke alert"
        stem_hpi = " last known well time two hours incontinence"
        stem_gen = " severe sensory loss"
        stem_pe = " stoke severity moderate"
        stem_dl = " cbc with diff no occlusion"
        stem_di = " internal carotid"
        stem_dx = " cerebral infarction aca"
        stem_tx = " asa seizure precautions dvt prophylaxis"
        stem_consult = " intensivist neurologist"
        stem_inter = " iv tpa"
        stem_found = " acute ischemic stroke aca symptoms"
        text = (
            stem_cc
            + stem_hpi
            + stem_gen
            + stem_pe
            + stem_dl
            + stem_di
            + stem_dx
            + stem_tx
            + stem_consult
            + stem_inter
            + stem_found
        )
        return text

    text = create_case()
    transcript = Transcript(
        type=Transcriber.DEEPGRAM, raw=text, word_count=len(text.split())
    )
    clean_txt = clean(transcript)
    case = Case(
        code="I63.43",
        system="Neruo",
        name="cerebral infarction aca < 3 hrs",
    )

    correct_soap, wrong_soap = get_keywords(clean_txt.clean, case)

    assert correct_soap is not None

    assert correct_soap.histories.totals.count == 4
    assert correct_soap.objectives.totals.count == 5
    assert correct_soap.assessments.totals.count == 1
    assert correct_soap.plans.totals.count == 6
    assert correct_soap.foundations.totals.count == 4

    assert correct_soap.totals.count == 20
    assert correct_soap.totals.expected == 36
    assert correct_soap.totals.word_count == 42

    assert wrong_soap is not None
    assert wrong_soap.totals.count == 0
    assert wrong_soap.totals.expected == 5
    assert wrong_soap.totals.word_count == 0



"""Testing for core module"""

import os

from audio_case_grade import Case, Transcriber, Transcript, clean, get_score, hello


def test_hello():
    """Test hello function"""

    result = hello()

    assert result == "Welcome to audio-case-grade!"


def test_score_no_density():
    """Test score function with no lexical density"""

    raw_text = "this is text"
    mock_clean_text = raw_text
    transcript = Transcript(
        type=Transcriber.DEEPGRAM, raw=raw_text, clean=mock_clean_text
    )

    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )
    expected_lexical_density = 0.0

    actual_score = get_score(transcript, case)

    assert actual_score.lexical_density == expected_lexical_density


def test_score_with_density():
    """Test score function with an easy to calculate lexical density"""

    raw_text = "hello world chest pain two months fatigued ago"
    mock_clean_text = raw_text
    transcript = Transcript(
        type=Transcriber.DEEPGRAM,
        raw=raw_text,
        clean=mock_clean_text,
    )
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )
    expected_lexical_density = 62.5

    actual_score = get_score(transcript, case)

    assert actual_score.lexical_density == expected_lexical_density


def test_score_with_clean():
    """Test score function"""

    raw_text = "hello world chest pain two months fatigued ago"
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=raw_text)
    transcript = clean(transcript)
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )
    expected_lexical_density = 62.5

    actual_score = get_score(transcript, case)

    assert actual_score.lexical_density == expected_lexical_density


def _get_transcript(trancript_type: str) -> str:
    current_directory = os.path.dirname(__file__)
    transcript_folder = os.path.join(current_directory, "sample_transcripts")

    transcript_name = trancript_type + " transcript.txt"
    transcript_path = os.path.join(transcript_folder, transcript_name)
    with open(transcript_path, "r", encoding="utf-8") as transcript_file:
        transcript = transcript_file.read()
    return transcript


def test_score_great_transcript():
    """Test score function for great transcript"""

    raw_text = _get_transcript("great")
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=raw_text)
    transcript = clean(transcript)
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )
    expected_lexical_density = 47.74

    actual_score = get_score(transcript, case)

    assert actual_score.lexical_density == expected_lexical_density


def test_score_average_transcript():
    """Test score function for average transcript"""

    raw_text = _get_transcript("average")
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=raw_text)
    transcript = clean(transcript)
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )
    expected_lexical_density = 38.41

    actual_score = get_score(transcript, case)

    assert actual_score.lexical_density == expected_lexical_density


def test_score_bad_transcript():
    """Test score function for bad transcript"""

    raw_text = _get_transcript("bad")
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=raw_text)
    transcript = clean(transcript)
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )
    expected_lexical_density = 23.59

    actual_score = get_score(transcript, case)

    assert actual_score.lexical_density == expected_lexical_density

    assert actual_score.lexical_density == expected_lexical_density



"""Module for textual Cleaning/Standardization"""

import importlib.resources as pkg_resources
import string
from typing import Mapping, Optional

import jiwer  # type: ignore
import nltk  # type: ignore
import pandas as pd
from nltk.corpus import stopwords  # type: ignore
from nltk.stem import WordNetLemmatizer  # type: ignore
from pandas import DataFrame

from .types import Transcriber, Transcript

ABBREVIATION_COLUMN = "Abbreviation"
MEANING_COLUMN = "Meaning"
STOP_WORDS_LANGUAGE = "english"
ACCEPTABLE_STOP_WORDS = ["no", "not", "most", "some", "out", "s", "t"]
COMMON_SUBSTITUTIONS: Mapping[str, str] = {
    "1+": "one plus",
    "1 +": "one plus",
    "2 d": "2d",
    "o two": "oxygen saturation",
    "o2": "oxygen saturation",
    "mg": "milligrams",
    "x-ray": "xray",
    "x ray": "xray",
    "cr td": "crtd",
    "rails": "rales",
    "rail": "rale",
}
UNACCEPTABLE_ABBREV = ["s", "f", "b", "l", "r", "w"]


def get_abbreviations_csv() -> DataFrame:
    """Get the medical abbreviations & acronyms from the csv file"""

    resources = "audio_case_grade.resources"
    abbreviations_file = "abbreviations.csv"

    with pkg_resources.open_text(resources, abbreviations_file) as csv_file:
        return pd.read_csv(csv_file, header=0)


def _get_abbreviation_map(
    abbreviations_override: Optional[DataFrame] = None,
) -> Mapping[str, str]:
    """Convert medical abbreviation & acronyms to map"""

    acr_abb = (
        abbreviations_override
        if abbreviations_override is not None
        else get_abbreviations_csv()
    )
    abbrev_map = dict(
        zip(
            acr_abb[ABBREVIATION_COLUMN].str.lower(),
            acr_abb[MEANING_COLUMN].str.lower(),
        )
    )
    for key in UNACCEPTABLE_ABBREV:
        abbrev_map.pop(key, None)
    return abbrev_map


def _text_clean(text: str, abbrev_map: Mapping[str, str]) -> str:
    """textual cleaning and standardization"""

    translator = str.maketrans(string.punctuation, " " * len(string.punctuation))
    t = jiwer.ExpandCommonEnglishContractions()(text)
    t = jiwer.RemoveKaldiNonWords()(t)
    t = t.translate(translator)
    t = jiwer.SubstituteWords(abbrev_map)(t)
    t = jiwer.SubstituteWords(COMMON_SUBSTITUTIONS)(t)
    t = jiwer.Strip()(t)
    clean_text = jiwer.RemoveMultipleSpaces()(t)
    return clean_text


def _load_stop_words() -> list[str]:
    """Load Stop Words"""

    nltk.download("stopwords", quiet=True)
    stop = stopwords.words(STOP_WORDS_LANGUAGE)
    for word in ACCEPTABLE_STOP_WORDS:
        stop.remove(word)
    return stop


def _load_lemmatizer() -> WordNetLemmatizer:
    nltk.download("wordnet", quiet=True)
    lemma = WordNetLemmatizer()
    return lemma


def _lemmatize(word: str, lemma: WordNetLemmatizer) -> str:
    lemma_word = lemma.lemmatize(lemma.lemmatize(word, pos="v"), pos="n")
    return lemma_word


def _internal_clean(
    raw_transcript: str, abbreviations_override: Optional[DataFrame] = None
) -> str:
    """Removes stop words from presentations, applies above cleaning functions"""
    lemma = _load_lemmatizer()
    stop = _load_stop_words()
    abbrev_map = _get_abbreviation_map(abbreviations_override)
    cleaned_text = _text_clean(raw_transcript, abbrev_map)
    clean_list = cleaned_text.split()
    clean_stop = [word for word in clean_list if word not in stop]
    clean_lemma = [_lemmatize(word, lemma) for word in clean_stop]
    return " ".join(clean_lemma)


def clean(
    transcript: Transcript, abbreviations_override: Optional[DataFrame] = None
) -> Transcript:
    """Clean up the transcript"""

    if transcript.type == Transcriber.DEEPGRAM:
        transcript.clean = _internal_clean(transcript.raw, abbreviations_override)
        return transcript

    if transcript.type == Transcriber.WHISPER:
        transcript.clean = _internal_clean(transcript.raw, abbreviations_override)
        return transcript

    return transcript



"""Tests for the clean module."""

from pandas import DataFrame

from audio_case_grade import Transcriber, Transcript, clean, get_abbreviations_csv


def test_stop_words():
    """Test clean function - Stop Words"""

    text = "This is the raw transcript"
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=text)
    expected_cleaned_text = "This raw transcript"

    cleaned_transcript = clean(transcript)

    assert cleaned_transcript.clean == expected_cleaned_text


def test_abbrev():
    """Test clean function - Abbreviations"""

    text = "wbc jvd bnp"
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=text)
    expected_cleaned_text = (
        "white blood cell jugular venous distention b-type natriuretic peptide"
    )

    cleaned_transcript = clean(transcript)

    assert cleaned_transcript.clean == expected_cleaned_text


def test_spaces():
    """Test clean function - Spacing Issues"""

    text = "  This is the   raw transcript "
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=text)
    expected_cleaned_text = "This raw transcript"

    cleaned_transcript = clean(transcript)

    assert cleaned_transcript.clean == expected_cleaned_text


def test_lemma():
    """Test Lemmatizer"""
    text = "the quick brown foxes are jumping over the lazy dogs"
    transcript = Transcript(type=Transcriber.DEEPGRAM, raw=text)
    expected_cleaned_text = "quick brown fox jump lazy dog"

    cleaned_transcript = clean(transcript)

    assert cleaned_transcript.clean == expected_cleaned_text


def test_whisper_type():
    """Test  Whisper type"""
    text = "the quick brown foxes are jumping over the lazy dogs"
    transcript = Transcript(type=Transcriber.WHISPER, raw=text)
    expected_cleaned_text = "quick brown fox jump lazy dog"

    cleaned_transcript = clean(transcript)

    assert cleaned_transcript.clean == expected_cleaned_text


def test_get_abbreviations_csv():
    """Test getting abbreviations csv"""

    csv = get_abbreviations_csv()

    assert csv is not None
    assert isinstance(csv, DataFrame)



"""Testing for keybank module"""

import numpy as np
import pandas as pd
import pytest
from pandas import DataFrame

from audio_case_grade import Case, get_keybank, get_keybank_csv


def test_keybank():
    """Test keybank function"""

    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )

    correct_keybank, wrong_keybank = get_keybank(case)

    assert correct_keybank is not None
    assert wrong_keybank is not None
    assert len(correct_keybank) == 15
    assert len(wrong_keybank) == 15


def test_get_keybank_for_abbreviations():
    """Test keybank for abbreviations"""

    case = Case(
        code="K80.00",
        system="GI",
        name="acute calculus cholecystitis",
    )

    correct_keybank, wrong_keybank = get_keybank(case)

    assert correct_keybank is not None
    assert wrong_keybank is not None
    assert correct_keybank.iloc[0] == ["abdominal pain"]
    assert "elevate white blood cell" in correct_keybank.iloc[7]
    assert wrong_keybank.iloc[0] == []


def test_get_keybank_for_contamination():
    """Test to ensure the wrong keybank doesn't contain the correct keywords"""

    case = Case(
        code="I25.2",
        system="Cardiopulm",
        name="ischemic cardiomyopathy",
    )

    correct_keybank, wrong_keybank = get_keybank(case)

    assert correct_keybank is not None
    assert wrong_keybank is not None
    assert sorted(correct_keybank.iloc[1]) == sorted(
        ["short breath", "six month", "physical activity"]
    )
    assert "nyha two" in wrong_keybank.iloc[9]
    assert "nyha four" not in wrong_keybank.iloc[9]
    assert "myocarditis dilate cardiomyopathy" in wrong_keybank.iloc[9]


def test_get_keybank_for_lemmatization():
    """Test to ensure the  keybank is Cleaned"""

    case = Case(
        code="I42.5",
        system="Cardiopulm",
        name="sarcoidosis r. cardiomyopathy",
    )

    correct_keybank, wrong_keybank = get_keybank(case)

    assert correct_keybank is not None
    assert wrong_keybank is not None
    assert sorted(correct_keybank.iloc[1]) == sorted(
        ["pre syncope", "physical activity", "palpitation", "short breath", "two month"]
    )
    assert "wheeze" in sorted(correct_keybank.iloc[2])


def test_get_keybank_csv():
    """Test getting keybank csv"""

    csv = get_keybank_csv()

    assert csv is not None
    assert isinstance(csv, DataFrame)


def test_dimensions():
    """Check Dimensions of Override DataFrames"""

    with pytest.raises(ValueError):
        case = Case(
            code="I42.1",
            system="Cardiopulm",
            name="concentric left ventricular hypertrophy",
        )
        num_rows = 1
        num_cols = 15

        data = {f"col{i}": ["chest pain"] * num_rows for i in range(1, num_cols + 1)}
        df = pd.DataFrame(data)
        get_keybank(case, keybank_override=df)


def test_empty_cell():
    """Check Empty Cell Remains Empty"""
    case = Case(
        code="I42.1",
        system="Cardiopulm",
        name="concentric left ventricular hypertrophy",
    )
    csv = get_keybank_csv().replace(r"^\s*$", np.nan, regex=True)
    correct_keybank, wrong_keybank = get_keybank(case, keybank_override=csv)

    assert correct_keybank.iloc[14] == []
    assert wrong_keybank.iloc[14] == []



"""Handle the KeyBank data"""

import importlib.resources as pkg_resources
from typing import Mapping, Optional, Tuple, cast

import pandas as pd
from jiwer import (  # type: ignore
    Compose,
    RemoveMultipleSpaces,
    RemoveSpecificWords,
    Strip,
    SubstituteWords,
)
from pandas import DataFrame, Series

from .clean import (
    _get_abbreviation_map,
    _lemmatize,
    _load_lemmatizer,
    _load_stop_words,
    get_abbreviations_csv,
)
from .types import Case

SEPERATOR = ","
CODE_COLUMN = "ICD10"
SYSTEM_COLUMN = "System"
CASE_COLUMNS = ["Case", CODE_COLUMN, SYSTEM_COLUMN]
EXPECTED_KEYBANK_COLS = 18
EXPECTED_ABBREV_COLS = 2


def get_keybank_csv() -> DataFrame:
    """Get the KeyBank from the csv file"""

    resources = "audio_case_grade.resources"
    keybank_file = "keybank.csv"

    with pkg_resources.open_text(resources, keybank_file) as csv_file:
        return pd.read_csv(
            csv_file,
            na_filter=False,
        )


def _check_for_dupes_keybank(
    keybank: DataFrame, ignore_error: Optional[bool] = False
) -> DataFrame:
    """Find all rows with duplicates"""

    def has_duplicates(row):
        seen = set()
        dupes = set()
        for col in keybank.columns:
            if col not in CASE_COLUMNS and isinstance(row[col], list):
                for item in row[col]:
                    if item in seen:
                        dupes.add(item)
                    else:
                        seen.add(item)
        has_dupes = len(dupes) > 0
        if has_dupes and not ignore_error:
            raise ValueError(
                f"Keybank case found with duplicate keywords: {row[CODE_COLUMN]} {dupes}"
            )
        return has_dupes

    return keybank[keybank.apply(has_duplicates, axis=1)]


def _check_for_dupes_wrong(keybank: Series) -> Series:
    """Check for Duplication in the Wrong Keybank"""
    seen = set()
    result = []

    for lst in keybank:
        new_lst = []
        for item in lst:
            if item not in seen:
                new_lst.append(item)
                seen.add(item)
        result.append(new_lst)

    return pd.Series(result, index=keybank.index)


def _clean_keybank(keybank: DataFrame, abbrev_map: Mapping[str, str]) -> DataFrame:
    stop = _load_stop_words()
    lemma = _load_lemmatizer()
    text_normalization = Compose(
        [
            RemoveSpecificWords(stop),
            SubstituteWords(abbrev_map),
            Strip(),
            RemoveMultipleSpaces(),
        ]
    )

    def clean_cell(cell):
        if pd.isna(cell):
            return []
        if isinstance(cell, str):
            cleaned = text_normalization(cell)
            cleaned_lemma = [
                " ".join(_lemmatize(word, lemma) for word in key.split())
                for key in cleaned.split(SEPERATOR)
            ]
            return list({item.strip() for item in cleaned_lemma if item != ""})
        return []

    cleaned_keybank = keybank.copy()

    for column in cleaned_keybank.columns:
        if column not in CASE_COLUMNS:
            cleaned_keybank[column] = cleaned_keybank[column].apply(clean_cell)

    _check_for_dupes_keybank(cleaned_keybank)

    return cleaned_keybank


def _get_rows_by_code(keybank: DataFrame, case: Case) -> DataFrame:
    """Get the rows that match the case"""

    return keybank[keybank[CODE_COLUMN] == case.code]


def _get_rows_by_system(keybank: DataFrame, case: Case) -> DataFrame:
    """Get the rows that match the system but not case"""

    return keybank[keybank[SYSTEM_COLUMN] == case.system]


def _combine_rows(rows: DataFrame) -> Series:
    """Combine rows into a single row"""

    single_row_df = rows.apply(
        lambda col: list(
            {item.strip() for row in col for item in row if isinstance(row, list)}
        )
    )

    # Single row is already a series but has wrong type
    series = cast(Series, single_row_df)
    series.name = "Keywords"

    return series


def _subtract_series(s1: Series, s2: Series) -> Series:
    """Subtract the string arrays in df2 from df1."""

    return s1.combine(s2, lambda cell1, cell2: list(set(cell1) - set(cell2)))


def _check_csv_columns(csv: DataFrame, min_col) -> bool:
    if len(csv.columns) < min_col:
        raise ValueError(
            "DataFrame has incorrect dimensions. Expected Number of Columns: "
            + f"({min_col}, got {len(csv.columns)})"
        )

    return True


def get_keybank(
    case: Case,
    keybank_override: Optional[DataFrame] = None,
    abbreviations_override: Optional[DataFrame] = None,
) -> Tuple[Series, Series]:
    """Get the KeyBank"""

    keybank = get_keybank_csv() if keybank_override is None else keybank_override
    _check_csv_columns(keybank, EXPECTED_KEYBANK_COLS)
    abbrev = (
        get_abbreviations_csv()
        if abbreviations_override is None
        else abbreviations_override
    )
    _check_csv_columns(abbrev, EXPECTED_ABBREV_COLS)
    abbrev_map = _get_abbreviation_map(abbrev)

    cleaned_keybank = _clean_keybank(keybank, abbrev_map)

    correct_rows = _get_rows_by_code(cleaned_keybank, case).drop(CASE_COLUMNS, axis=1)
    correct_row = _combine_rows(correct_rows)

    wrong_rows = _get_rows_by_system(cleaned_keybank, case).drop(CASE_COLUMNS, axis=1)
    wrong_row = _combine_rows(wrong_rows)
    wrong_row_without_correct = _subtract_series(wrong_row, correct_row)
    wrong_row_without_dupes = _check_for_dupes_wrong(wrong_row_without_correct)
    return correct_row, wrong_row_without_dupes



"""Tests Lexical Density Grade Scale"""

from audio_case_grade import get_grade


def test_get_grade():
    """Testing Lexical Density Scale"""
    actual_1 = get_grade(45)
    actual_2 = get_grade(35)
    actual_3 = get_grade(25)
    actual_4 = get_grade(15)
    assert actual_1 == "Advanced"
    assert actual_2 == "Proficient"
    assert actual_3 == "Average"
    assert actual_4 == "Low"
